#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import math
import numpy as np

import matplotlib
from matplotlib import cm
import matplotlib.pyplot as plt
from scipy import misc 

import GridMap as gmap

PROBAB_FREE_MAX = 0.4
PROBAB_OCC_MIN = 0.6

IS_FRONTIER_KEY = 1
IS_NOT_FRONTIER_KEY = 0

class ExplorationMap(gmap.GridMap):

    def __init__(self, width=0, height=0, scale=0.1):
        """
        Parameters
        ----------
        width : int, optional
            width of the grid map
        height : int, optional
            height of the grid map
        scale : double, optional
            the scale of the map 
        """
        if width > 0 and height > 0:
            self.grid = np.zeros((width, height), dtype=[('p',np.dtype(float)),('free',np.dtype(bool)),('frontier',np.dtype(int))])
            self.grid['p'] = 0.5    #set all probabilities to 0.5
            self.grid['free'] = True #set all cells to passable
            self.grid['frontier'] = IS_NOT_FRONTIER_KEY
            self.width = width
            self.height = height
            self.scale = scale
        else:
            self.grid = None
            self.width = None
            self.height = None
            self.scale = None

    ##################################################
    # Image and out txt helper functions
    ##################################################

    def load_map_image(self, filename, scale=1):
        image = misc.imread(filename)
        self.width = image.shape[0]
        self.height = image.shape[1]
        self.scale = scale
        self.grid = np.zeros((self.width,self.height), dtype=[('p',np.dtype(float)),('free',np.dtype(bool)),('frontier',np.dtype(int))])
        for x in range(self.width):
            for y in range(self.height):
                self.set_cell_p( (x,y), (3*255-(float(image[x,y,0])+float(image[x,y,1])+float(image[x,y,2])))/(3*255) )
        self.detect_frontiers()
        return image

    def draw_frontiers_to_image(self, imageIn, imageOut, scale=1 ):
        image = self.load_map_image( imageIn, scale )
        for x in range(1,self.width-1):
            for y in range(1,self.height-1):
                if self.grid['frontier'][x, y] == IS_FRONTIER_KEY:
                    image[x,y,0] = 255
                    image[x,y,1] = 0
                    image[x,y,2] = 0
                    image[x,y,3] = 255
        misc.imsave(imageOut, image)
    
    def save_frontiers_to_file(self, outFile ):
        frontiers = self.list_frontiers()
        with open( outFile, 'w') as f:
            for x in range(len(frontiers)):
                f.write("%d" % frontiers[x][0] )
                f.write(" %d\n" % frontiers[x][1] )

    ##############################################
    ## Helper functions for exploration
    ##############################################
    def get_probability(self, x, y):
        return self.get_cell_p((x,y))
        
    def isKnownFree(self, x, y):
        return self.get_probability(x,y)<=PROBAB_FREE_MAX
    
    def isKnownOccupied(self, x, y):
        return self.get_probability(x,y)>=PROBAB_OCC_MIN
        
    def isUnknown(self, x, y):
        return ~self.isKnownFree( x, y) and ~self.isKnownOccupied( x, y)
        
    def detect_frontiers(self):
        ret = []
        for x in range(1,self.width-1):
            for y in range(1,self.height-1):
                if self.is_frontier(x, y):
                    self.grid['frontier'][x,y] = IS_FRONTIER_KEY

    def list_frontiers(self):               
        ret = []
        for x in range(1,self.width-1):
            for y in range(1,self.height-1):
                if self.grid['frontier'][x,y] == IS_FRONTIER_KEY:
                    ret.append((x,y))
        return ret            
        
    def is_frontier(self, x, y):
        """
        Method to determine whether a cell is a frontier.
    
        Parameters
        ----------
        x : int
            x coordinate of the cell
        y : int
            y coordinate of the cell
        ----------

        Returns
        ----------
        boolean
            True if the cell is a frontier, False otherwise 
        """

        # TODO t1e-expl

        if (x + y) % 2 == 0:
            return False
        return True

##############################################
## Helper functions to plot the map
##############################################

def plot_frontiers(emap):
    grid = emap.grid['p'].transpose()
    norm = matplotlib.colors.Normalize(vmin=0, vmax=1)
    frontiers = emap.list_frontiers()
    cols = cm.binary(norm(grid))
    for pt in frontiers:
        cols[pt[1],pt[0],:3] = [1,0,0]
    plt.xlabel('x [grid cells]')
    plt.ylabel('y [grid cells]')
    plt.imshow(cols, interpolation='nearest', vmin=0, vmax=1)
    plt.draw()
    
